@@ -17,8 +17,17 @@
  * Boston, MA 02111-1307, USA.
  */
 
+/*
+ * Modified by the GTK+ Team and others 1997-1999.  See the AUTHORS
+ * file for a list of people on the GTK+ Team.  See the ChangeLog
+ * files for a list of changes.  These files are distributed with
+ * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
+ */
+
 #if defined(XINPUT_GXI) || defined(XINPUT_XFREE)
 
+#include <string.h>
+
 /* Forward declarations */
 static void gdk_input_get_root_relative_geometry (Display *dpy, Window w, 
 						  int *x_ret, int *y_ret,
@@ -162,6 +171,7 @@ gdk_input_device_new(XDeviceInfo *device, gint include_core)
 
   gdkdev->info.num_axes = 0;
   gdkdev->info.num_keys = 0;
+  gdkdev->info.axes = NULL;
   gdkdev->info.keys = NULL;
   gdkdev->axes = 0;
   gdkdev->info.has_cursor = 0;
@@ -253,6 +263,8 @@ gdk_input_device_new(XDeviceInfo *device, gint include_core)
 	g_free(gdkdev->axes);
       if (gdkdev->info.keys)
 	g_free(gdkdev->info.keys);
+      if (gdkdev->info.axes)
+	g_free (gdkdev->info.axes);
       g_free(gdkdev);
       return NULL;
     }
@@ -436,20 +448,22 @@ gdk_input_common_init(gint include_core)
   for (loop = 0; loop < num_extensions &&
 	 (strcmp(extensions[loop], "XInputExtension") != 0); loop++);
   XFreeExtensionList(extensions);
-  if (loop == num_extensions)	/* XInput extension not found */
-    return FALSE;
-
-  gdk_input_devices = 0;
-  devices = XListInputDevices(display, &num_devices);
-  
-  for(loop=0; loop<num_devices; loop++)
+  gdk_input_devices = NULL;
+  if (loop < num_extensions)
     {
-      GdkDevicePrivate *gdkdev = gdk_input_device_new(&devices[loop],
+      /* XInput extension found */
+
+      devices = XListInputDevices(display, &num_devices);
+  
+      for(loop=0; loop<num_devices; loop++)
+	{
+	  GdkDevicePrivate *gdkdev = gdk_input_device_new(&devices[loop],
 						      include_core);
-      if (gdkdev)
-	gdk_input_devices = g_list_append(gdk_input_devices, gdkdev);
+	  if (gdkdev)
+	    gdk_input_devices = g_list_append(gdk_input_devices, gdkdev);
+	}
+      XFreeDeviceList(devices);
     }
-  XFreeDeviceList(devices);
 
   gdk_input_devices = g_list_append (gdk_input_devices, &gdk_input_core_info);
 
@@ -877,6 +891,8 @@ gdk_input_common_get_pointer     (GdkWindow       *window,
 	    }
 	  input_class = (XInputClass *)(((char *)input_class)+input_class->length);
 	}
+
+      XFreeDeviceState (state); 
     }
 }
 
