@@ -34,17 +34,9 @@
 #include "gdkprivate.h"
 #include "MwmUtil.h"
 
-#if HAVE_CONFIG_H
-#  include <config.h>
-#  if STDC_HEADERS
-#    include <stdlib.h>
-#    include <stdio.h>
-#    include <string.h>
-#  endif
-#else
-#  include <stdlib.h>
-#  include <stdio.h>
-#endif
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
 
 
 #ifdef HAVE_SHAPE_EXT
@@ -409,7 +401,7 @@ gdk_window_new (GdkWindow     *parent,
     {
       depth = 0;
       class = InputOnly;
-      private->colormap = NULL;
+      private->colormap = gdk_colormap_get_system ();
     }
   
   private->xwindow = XCreateWindow (private->xdisplay, xparent,
@@ -707,10 +699,10 @@ gdk_window_destroy_notify (GdkWindow *window)
   
   if (!private->destroyed)
     {
-      if (private->window_type == GDK_WINDOW_FOREIGN)
-	gdk_window_internal_destroy (window, FALSE, FALSE);
-      else
+      if (private->window_type != GDK_WINDOW_FOREIGN)
 	g_warning ("GdkWindow %#lx unexpectedly destroyed", private->xwindow);
+
+      gdk_window_internal_destroy (window, FALSE, FALSE);
     }
   
   gdk_xid_table_remove (private->xwindow);
@@ -732,6 +724,7 @@ gdk_window_unref (GdkWindow *window)
 {
   GdkWindowPrivate *private = (GdkWindowPrivate *)window;
   g_return_if_fail (window != NULL);
+  g_return_if_fail (private->ref_count > 0);
   
   private->ref_count -= 1;
   if (private->ref_count == 0)
@@ -1089,8 +1082,10 @@ gdk_window_set_hints (GdkWindow *window,
       size_hints.max_height = max_height;
     }
   
-  if (flags)
-    XSetWMNormalHints (private->xdisplay, private->xwindow, &size_hints);
+  /* FIXME: Would it be better to delete this property of
+   *        flags == 0? It would save space on the server
+   */
+  XSetWMNormalHints (private->xdisplay, private->xwindow, &size_hints);
 }
 
 void 
@@ -1110,7 +1105,15 @@ gdk_window_set_geometry_hints (GdkWindow      *window,
   size_hints.flags = 0;
   
   if (geom_mask & GDK_HINT_POS)
-    size_hints.flags |= PPosition;
+    {
+      size_hints.flags |= PPosition;
+      /* We need to initialize the following obsolete fields because KWM 
+       * apparently uses these fields if they are non-zero.
+       * #@#!#!$!.
+       */
+      size_hints.x = 0;
+      size_hints.y = 0;
+    }
   
   if (geom_mask & GDK_HINT_MIN_SIZE)
     {
@@ -1145,28 +1148,30 @@ gdk_window_set_geometry_hints (GdkWindow      *window,
       size_hints.flags |= PAspect;
       if (geometry->min_aspect <= 1)
 	{
-	  size_hints.min_aspect.x = G_MAXINT * geometry->min_aspect;
-	  size_hints.min_aspect.y = G_MAXINT;
+	  size_hints.min_aspect.x = 65536 * geometry->min_aspect;
+	  size_hints.min_aspect.y = 65536;
 	}
       else
 	{
-	  size_hints.min_aspect.x = G_MAXINT;
-	  size_hints.min_aspect.y = G_MAXINT / geometry->min_aspect;;
+	  size_hints.min_aspect.x = 65536;
+	  size_hints.min_aspect.y = 65536 / geometry->min_aspect;;
 	}
       if (geometry->max_aspect <= 1)
 	{
-	  size_hints.max_aspect.x = G_MAXINT * geometry->max_aspect;
-	  size_hints.max_aspect.y = G_MAXINT;
+	  size_hints.max_aspect.x = 65536 * geometry->max_aspect;
+	  size_hints.max_aspect.y = 65536;
 	}
       else
 	{
-	  size_hints.max_aspect.x = G_MAXINT;
-	  size_hints.max_aspect.y = G_MAXINT / geometry->max_aspect;;
+	  size_hints.max_aspect.x = 65536;
+	  size_hints.max_aspect.y = 65536 / geometry->max_aspect;;
 	}
     }
-  
-  if (geom_mask)
-    XSetWMNormalHints (private->xdisplay, private->xwindow, &size_hints);
+
+  /* FIXME: Would it be better to delete this property of
+   *        geom_mask == 0? It would save space on the server
+   */
+  XSetWMNormalHints (private->xdisplay, private->xwindow, &size_hints);
 }
 
 void
@@ -1235,7 +1240,7 @@ gdk_window_set_background (GdkWindow *window,
 void
 gdk_window_set_back_pixmap (GdkWindow *window,
 			    GdkPixmap *pixmap,
-			    gint       parent_relative)
+			    gboolean   parent_relative)
 {
   GdkWindowPrivate *window_private;
   GdkPixmapPrivate *pixmap_private;
@@ -1988,7 +1993,7 @@ gdk_window_remove_filter (GdkWindow     *window,
 	  if (private)
 	    private->filters = g_list_remove_link (private->filters, node);
 	  else
-	    gdk_default_filters = g_list_remove_link (gdk_default_filters, tmp_list);
+	    gdk_default_filters = g_list_remove_link (gdk_default_filters, node);
 	  g_list_free_1 (node);
 	  g_free (filter);
 	  
@@ -2061,8 +2066,8 @@ gdk_window_set_icon (GdkWindow *window,
 }
 
 void          
-gdk_window_set_icon_name (GdkWindow *window, 
-			  gchar *    name)
+gdk_window_set_icon_name (GdkWindow   *window, 
+			  const gchar *name)
 {
   GdkWindowPrivate *window_private;
   XTextProperty property;
@@ -2073,7 +2078,7 @@ gdk_window_set_icon_name (GdkWindow *window,
   if (window_private->destroyed)
     return;
   res = XmbTextListToTextProperty (window_private->xdisplay,
-				   &name, 1, XStdICCTextStyle,
+				   (char **)&name, 1, XStdICCTextStyle,
                                	   &property);
   if (res < 0)
     {
@@ -2206,6 +2211,8 @@ gdk_window_get_toplevels (void)
   return new_list;
 }
 
+#ifdef HAVE_SHAPE_EXT
+
 /* 
  * propagate the shapes from all child windows of a GDK window to the parent 
  * window. Shamelessly ripped from Enlightenment's code
@@ -2519,6 +2526,8 @@ gdk_propagate_shapes (Display *disp,
   g_free (spans);
 }
 
+#endif /* HAVE_SHAPE_EXT */
+
 void
 gdk_window_set_child_shapes (GdkWindow *window)
 {
