@@ -36,6 +36,8 @@
 
 #define MIN_ENTRY_WIDTH  150
 #define DRAW_TIMEOUT     20
+
+/* If you are going to change this, see the note in entry_adjust_scroll */
 #define INNER_BORDER     2
 
 enum {
@@ -372,6 +374,48 @@ gtk_entry_new (void)
   return GTK_WIDGET (gtk_type_new (GTK_TYPE_ENTRY));
 }
 
+static GdkWChar
+gtk_entry_get_invisible_char (GtkEntry *entry)
+{
+  GdkWChar ch = 0;
+
+  if (entry->use_wchar)
+    gdk_mbstowcs (&ch, "*", 1);
+  else
+    ch = '*';
+
+  return ch;
+}
+
+/*
+ * Draws the string, noting that if entry->use_wchar is false, then
+ * the text is not really wide characters, but narrow characters
+ * stored as wide characters.
+ */
+static void
+gtk_entry_draw_wchars (GtkEntry       *entry,
+		       GdkDrawable    *drawable,
+		       GdkFont	      *font,
+		       GdkGC	      *gc,
+		       gint	       x,
+		       gint	       y,
+		       const GdkWChar *text,
+		       gint	       text_length)
+{
+  if (entry->use_wchar)
+    gdk_draw_text_wc (drawable, font, gc, x, y, text, text_length);
+  else
+    {
+      gint i;
+      gchar *mbstr = g_new (gchar, text_length);
+      
+      for (i = 0; i < text_length; i++)
+        mbstr[i] = text[i];
+      gdk_draw_text (drawable, font, gc, x, y, mbstr, text_length);
+      g_free(mbstr);
+    }
+}
+
 GtkWidget*
 gtk_entry_new_with_max_length (guint16 max)
 {
@@ -907,6 +951,8 @@ gtk_entry_button_press (GtkWidget      *widget,
 	default:
 	  break;
 	}
+
+      return TRUE;
     }
   else if (event->type == GDK_BUTTON_PRESS)
     {
@@ -930,6 +976,8 @@ gtk_entry_button_press (GtkWidget      *widget,
 	  if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) == widget->window)
 	    gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY, event->time);
 	}
+
+      return TRUE;
     }
 
   return FALSE;
@@ -973,10 +1021,14 @@ gtk_entry_button_release (GtkWidget      *widget,
 	  if (gdk_selection_owner_get (GDK_SELECTION_PRIMARY) == widget->window)
 	    gtk_selection_owner_set (NULL, GDK_SELECTION_PRIMARY, event->time);
 	}
+
+      return TRUE;
     }
   else if (event->button == 3)
     {
       gtk_grab_remove (widget);
+
+      return TRUE;
     }
 
   return FALSE;
@@ -1007,7 +1059,7 @@ gtk_entry_motion_notify (GtkWidget      *widget,
   entry_adjust_scroll (entry);
   gtk_entry_queue_draw (entry);
 
-  return FALSE;
+  return TRUE;
 }
 
 static gint
@@ -1101,21 +1153,39 @@ gtk_entry_key_press (GtkWidget   *widget,
       break;
     case GDK_Left:
       return_val = TRUE;
-      if (event->state & GDK_CONTROL_MASK)
-	gtk_move_backward_word (entry);
+      if (!extend_selection &&
+	  editable->selection_start_pos != editable->selection_end_pos)
+	{
+	  editable->current_pos = MIN (editable->selection_start_pos, editable->selection_end_pos);
+	  initial_pos = (guint)-1; /* Force redraw below */
+	}
       else
-	gtk_move_backward_character (entry);
+	{
+	  if (event->state & GDK_CONTROL_MASK)
+	    gtk_move_backward_word (entry);
+	  else
+	    gtk_move_backward_character (entry);
+	}
       break;
     case GDK_Right:
       return_val = TRUE;
-      if (event->state & GDK_CONTROL_MASK)
-	gtk_move_forward_word (entry);
+      if (!extend_selection &&
+	  editable->selection_start_pos != editable->selection_end_pos)
+	{
+	  editable->current_pos = MAX (editable->selection_start_pos, editable->selection_end_pos);
+	  initial_pos = (guint)-1; /* Force redraw below */
+	}
       else
-	gtk_move_forward_character (entry);
+	{
+	  if (event->state & GDK_CONTROL_MASK)
+	    gtk_move_forward_word (entry);
+	  else
+	    gtk_move_forward_character (entry);
+	}
       break;
     case GDK_Return:
       return_val = TRUE;
-      gtk_signal_emit_by_name (GTK_OBJECT (entry), "activate");
+      gtk_widget_activate (widget);
       break;
     /* The next two keys should not be inserted literally. Any others ??? */
     case GDK_Tab:
@@ -1370,19 +1440,20 @@ gtk_entry_draw_text (GtkEntry *entry)
       else
 	{
 	  gint i;
+	  GdkWChar invisible_char = gtk_entry_get_invisible_char (entry);
 	  
 	  stars = g_new (GdkWChar, end_pos - start_pos);
 	  for (i = 0; i < end_pos - start_pos; i++)
-	    stars[i] = '*';
+	    stars[i] = invisible_char;
 	  toprint = stars;
 	}
       
       if (selection_start_pos > start_pos)
-	gdk_draw_text_wc (drawable, widget->style->font,
-			  widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
-			  INNER_BORDER + start_xoffset, y,
-			  toprint,
-			  selection_start_pos - start_pos);
+	gtk_entry_draw_wchars (entry, drawable, widget->style->font,
+			       widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
+			       INNER_BORDER + start_xoffset, y,
+			       toprint,
+			       selection_start_pos - start_pos);
       
       if ((selection_end_pos >= start_pos) && 
 	  (selection_start_pos < end_pos) &&
@@ -1395,19 +1466,19 @@ gtk_entry_draw_text (GtkEntry *entry)
 				INNER_BORDER,
 				selection_end_xoffset - selection_start_xoffset,
 				height - 2*INNER_BORDER);
-	    gdk_draw_text_wc (drawable, widget->style->font,
-			      widget->style->fg_gc[selected_state],
-			      INNER_BORDER + selection_start_xoffset, y,
-			      toprint + selection_start_pos - start_pos,
-			      selection_end_pos - selection_start_pos);
+	    gtk_entry_draw_wchars (entry, drawable, widget->style->font,
+				   widget->style->fg_gc[selected_state],
+				   INNER_BORDER + selection_start_xoffset, y,
+				   toprint + selection_start_pos - start_pos,
+				   selection_end_pos - selection_start_pos);
 	 }	    
        
        if (selection_end_pos < end_pos)
-	 gdk_draw_text_wc (drawable, widget->style->font,
-			   widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
-			   INNER_BORDER + selection_end_xoffset, y,
-			   toprint + selection_end_pos - start_pos,
-			   end_pos - selection_end_pos);
+	 gtk_entry_draw_wchars (entry, drawable, widget->style->font,
+				widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
+				INNER_BORDER + selection_end_xoffset, y,
+				toprint + selection_end_pos - start_pos,
+				end_pos - selection_end_pos);
        /* free the space allocated for the stars if it's neccessary. */
       if (!editable->visible)
 	g_free (toprint);
@@ -1477,10 +1548,15 @@ gtk_entry_draw_cursor_on_drawable (GtkEntry *entry, GdkDrawable *drawable)
 	   */
 	  if ((editable->current_pos < entry->text_length) &&
 	      (editable->selection_start_pos == editable->selection_end_pos))
-	    gdk_draw_text_wc (drawable, widget->style->font,
-			      widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
-			      xoffset, yoffset,
-			      entry->text + editable->current_pos, 1);
+	    {
+	      GdkWChar c = editable->visible ?
+		                 *(entry->text + editable->current_pos) :
+		                 '*';
+	      
+	      gtk_entry_draw_wchars (entry, drawable, widget->style->font,
+				     widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
+				     xoffset, yoffset, &c, 1);
+	    }
 	}
 
 
@@ -1575,6 +1651,7 @@ entry_adjust_scroll (GtkEntry *entry)
     return;
 
   gdk_window_get_size (entry->text_area, &text_area_width, NULL);
+  text_area_width -= 2 * INNER_BORDER;
 
   /* Display as much text as we can */
   max_offset = MAX(0, entry->char_offset[entry->text_length] - text_area_width);
@@ -1582,14 +1659,23 @@ entry_adjust_scroll (GtkEntry *entry)
   if (entry->scroll_offset > max_offset)
     entry->scroll_offset = max_offset;
 
-  /* And make sure cursor is on screen */
+  /* And make sure cursor is on screen. Note that the cursor is
+   * actually drawn one pixel into the INNER_BORDER space on
+   * the right, when the scroll is at the utmost right. This
+   * looks better to to me than confining the cursor inside the
+   * border entirely, though it means that the cursor gets one
+   * pixel closer to the the edge of the widget on the right than
+   * on the left. This might need changing if one changed
+   * INNER_BORDER from 2 to 1, as one would do on a
+   * small-screen-real-estate display.
+   */
   xoffset = entry->char_offset[GTK_EDITABLE(entry)->current_pos];
   xoffset -= entry->scroll_offset;
 
   if (xoffset < 0)
     entry->scroll_offset += xoffset;
   else if (xoffset > text_area_width)
-    entry->scroll_offset += xoffset - text_area_width + 1;
+    entry->scroll_offset += xoffset - text_area_width;
 
   gtk_widget_queue_draw (GTK_WIDGET (entry));
 }
@@ -1737,21 +1823,23 @@ gtk_entry_insert_text (GtkEditable *editable,
       gint offset = 0;
       
       for (i = last_pos; i >= end_pos; i--)
-	entry->char_offset[i] 
-	  = entry->char_offset[i - insertion_length];
+	entry->char_offset[i] = entry->char_offset[i - insertion_length];
       
       for (i=start_pos; i<end_pos; i++)
 	{
+	  GdkWChar ch;
+
 	  entry->char_offset[i] = entry->char_offset[start_pos] + offset;
+
 	  if (editable->visible)
-	    {
-	      offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font,
-					   entry->text[i]);
-	    }
+	    ch = entry->text[i];
+	  else 
+	    ch = gtk_entry_get_invisible_char (entry);
+
+	  if (entry->use_wchar)
+	    offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font, ch);
 	  else
-	    {
-	      offset += gdk_char_width (GTK_WIDGET (entry)->style->font, '*');
-	    }
+	    offset += gdk_char_width (GTK_WIDGET (entry)->style->font, ch);
 	}
       for (i = end_pos; i <= last_pos; i++)
 	entry->char_offset[i] += offset;
@@ -1770,21 +1858,25 @@ gtk_entry_recompute_offsets (GtkEntry *entry)
 {
   gint i;
   gint offset = 0;
+  GtkEditable *editable = GTK_EDITABLE (entry);
 
   for (i=0; i<entry->text_length; i++)
     {
-      entry->char_offset[i] = offset;
-      if (GTK_EDITABLE (entry)->visible)
-	{
-	  offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font,
-				       entry->text[i]);
-	}
-      else
-	{
-	  offset += gdk_char_width (GTK_WIDGET (entry)->style->font, '*');
-	}
-    }
+      GdkWChar ch;
 
+      entry->char_offset[i] = offset;
+
+      if (editable->visible)
+	ch = entry->text[i];
+      else
+	ch = gtk_entry_get_invisible_char (entry);
+
+      if (entry->use_wchar)
+	offset += gdk_char_width_wc (GTK_WIDGET (entry)->style->font, ch);
+      else
+	offset += gdk_char_width (GTK_WIDGET (entry)->style->font, ch);
+    }
+  
   entry->char_offset[i] = offset;
 }
 
@@ -1932,10 +2024,16 @@ static void
 gtk_entry_move_word (GtkEditable *editable,
 		     gint         n)
 {
-  while (n-- > 0)
-    gtk_move_forward_word (GTK_ENTRY (editable));
-  while (n++ < 0)
-    gtk_move_backward_word (GTK_ENTRY (editable));
+  while (n > 0)
+    {
+      gtk_move_forward_word (GTK_ENTRY (editable));
+      n--;
+    }
+  while (n < 0)
+    {
+      gtk_move_backward_word (GTK_ENTRY (editable));
+      n++;
+    }
 }
 
 static void
@@ -1947,6 +2045,13 @@ gtk_move_forward_word (GtkEntry *entry)
 
   editable = GTK_EDITABLE (entry);
 
+  /* Prevent any leak of information */
+  if (!editable->visible)
+    {
+      editable->current_pos = entry->text_length;
+      return;
+    }
+
   if (entry->text && (editable->current_pos < entry->text_length))
     {
       text = entry->text;
@@ -1978,6 +2083,13 @@ gtk_move_backward_word (GtkEntry *entry)
 
   editable = GTK_EDITABLE (entry);
 
+  /* Prevent any leak of information */
+  if (!editable->visible)
+    {
+      editable->current_pos = 0;
+      return;
+    }
+
   if (entry->text && editable->current_pos > 0)
     {
       text = entry->text;
@@ -2168,14 +2280,16 @@ gtk_entry_set_selection (GtkEditable       *editable,
 			 gint               start,
 			 gint               end)
 {
+  gint length = GTK_ENTRY (editable)->text_length;
+  
   g_return_if_fail (editable != NULL);
   g_return_if_fail (GTK_IS_ENTRY (editable));
 
   if (end < 0)
-    end = GTK_ENTRY (editable)->text_length;
+    end = length;
   
-  editable->selection_start_pos = start;
-  editable->selection_end_pos = end;
+  editable->selection_start_pos = CLAMP (start, 0, length);
+  editable->selection_end_pos = MIN (end, length);
 
   gtk_entry_queue_draw (GTK_ENTRY (editable));
 }
