@@ -24,6 +24,7 @@
  * GTK+ at ftp://ftp.gtk.org/pub/gtk/. 
  */
 
+#include <config.h>
 #include <X11/Xlocale.h>	/* so we get the right setlocale */
 #include <ctype.h>
 #include <unistd.h>
@@ -33,6 +34,9 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#ifdef HAVE_CODESET
+#include <langinfo.h>
+#endif
 
 #include "gtkrc.h"
 #include "gtkbindings.h"
@@ -42,6 +46,7 @@
 typedef struct _GtkRcSet    GtkRcSet;
 typedef struct _GtkRcNode   GtkRcNode;
 typedef struct _GtkRcFile   GtkRcFile;
+typedef struct _GtkRcStylePrivate  GtkRcStylePrivate;
 
 struct _GtkRcSet
 {
@@ -57,6 +62,15 @@ struct _GtkRcFile
   gboolean reload;
 };
 
+struct _GtkRcStylePrivate
+{
+  GtkRcStyle style;
+
+  guint ref_count;
+  /* list of RC style lists including this RC style */
+  GSList *rc_style_lists;
+};
+
 static guint	   gtk_rc_style_hash		   (const char   *name);
 static gint	   gtk_rc_style_compare		   (const char   *a,
 						    const char   *b);
@@ -106,7 +120,6 @@ static void        gtk_rc_clear_hash_node          (gpointer   key,
 static void        gtk_rc_clear_styles               (void);
 static void        gtk_rc_append_default_pixmap_path (void);
 static void        gtk_rc_append_default_module_path (void);
-static void        gtk_rc_append_pixmap_path         (gchar *dir);
 static void        gtk_rc_add_initial_default_files  (void);
 
 
@@ -202,6 +215,11 @@ static gchar *pixmap_path[GTK_RC_MAX_PIXMAP_PATHS];
 #define GTK_RC_MAX_MODULE_PATHS 128
 static gchar *module_path[GTK_RC_MAX_MODULE_PATHS];
 
+/* A stack of directories for RC files we are parsing currently.
+ * these are implicitely added to the end of PIXMAP_PATHS
+ */
+GSList *rc_dir_stack = NULL;
+
 /* The files we have parsed, to reread later if necessary */
 GSList *rc_files = NULL;
 
@@ -233,7 +251,7 @@ gtk_rc_get_module_dir(void)
   if (var)
     path = g_strdup_printf("%s%s", var, "/lib/gtk/themes/engines");
   else
-    path = g_strdup_printf("%s%s", GTK_EXE_PREFIX, "/lib/gtk/themes/engines");
+    path = g_strdup_printf("%s%s", GTK_LIBDIR, "/gtk/themes/engines");
 
   return path;
 }
@@ -252,21 +270,12 @@ gtk_rc_append_default_pixmap_path(void)
   
   for (n = 0; pixmap_path[n]; n++) ;
   if (n >= GTK_RC_MAX_PIXMAP_PATHS - 1)
-    return;
-  pixmap_path[n++] = g_strdup(path);
-  pixmap_path[n] = NULL;
-  g_free(path);
-}
-
-static void
-gtk_rc_append_pixmap_path(gchar *dir)
-{
-  gint n;
-
-  for (n = 0; pixmap_path[n]; n++) ;
-  if (n >= GTK_RC_MAX_MODULE_PATHS - 1)
-    return;
-  pixmap_path[n++] = g_strdup(dir);
+    {
+      g_free (path);
+      return;
+    }
+  
+  pixmap_path[n++] = path;
   pixmap_path[n] = NULL;
 }
 
@@ -277,23 +286,20 @@ gtk_rc_append_default_module_path(void)
   gint n;
 
   for (n = 0; module_path[n]; n++) ;
-  if (n >= GTK_RC_MAX_MODULE_PATHS - 1)
+  if (n >= GTK_RC_MAX_MODULE_PATHS - 2)
     return;
   
   var = getenv("GTK_EXE_PREFIX");
   if (var)
     path = g_strdup_printf("%s%s", var, "/lib/gtk/themes/engines");
   else
-    path = g_strdup_printf("%s%s", GTK_EXE_PREFIX, "/lib/gtk/themes/engines");
-  module_path[n++] = g_strdup(path);
-  g_free(path);
+    path = g_strdup_printf("%s%s", GTK_LIBDIR, "/gtk/themes/engines");
+  module_path[n++] = path;
 
-  var = getenv("HOME");
+  var = g_get_home_dir ();
   if (var)
-    path = g_strdup_printf("%s%s", var, ".gtk/lib/themes/engines");
-  module_path[n++] = g_strdup(path);
+    module_path[n++] = g_strdup_printf("%s%s", var, "/.gtk/lib/themes/engines");
   module_path[n] = NULL;
-  g_free(path);
 }
 
 static void
@@ -328,9 +334,13 @@ gtk_rc_add_initial_default_files (void)
       gtk_rc_add_default_file (str);
       g_free (str);
 
-      str = g_strdup_printf ("%s%s", g_get_home_dir (), "/.gtkrc");
-      gtk_rc_add_default_file (str);
-      g_free (str);
+      var = g_get_home_dir ();
+      if (var)
+	{
+	  str = g_strdup_printf ("%s%s", var, "/.gtkrc");
+	  gtk_rc_add_default_file (str);
+	  g_free (str);
+	}
     }
 }
 
@@ -382,51 +392,173 @@ gtk_rc_get_default_files (void)
   return gtk_rc_default_files;
 }
 
+/* The following routine is based on _nl_normalize_codeset from
+ * the GNU C library. Contributed by
+ *
+ * Contributed by Ulrich Drepper <drepper@gnu.ai.mit.edu>, 1995.
+ * Copyright (C) 1995, 1996, 1997, 1998, 1999 Free Software Foundation, Inc.
+ * 
+ * Normalize codeset name.  There is no standard for the codeset
+ * names.  Normalization allows the user to use any of the common
+ * names.
+ */
+static char *
+_gtk_normalize_codeset (const char *codeset, int name_len)
+{
+  int len = 0;
+  int only_digit = 1;
+  char *retval;
+  char *wp;
+  int cnt;
+
+  for (cnt = 0; cnt < name_len; ++cnt)
+    if (isalnum (codeset[cnt]))
+      {
+	++len;
+
+	if (isalpha (codeset[cnt]))
+	  only_digit = 0;
+      }
+
+  retval = g_malloc ((only_digit ? 3 : 0) + len + 1);
+
+  if (only_digit)
+    {
+      strcpy (retval, "iso");
+      wp = retval + 3;
+    }
+  else
+    wp = retval;
+  
+  for (cnt = 0; cnt < name_len; ++cnt)
+    if (isalpha (codeset[cnt]))
+      *wp++ = isupper(codeset[cnt]) ? tolower (codeset[cnt]) : codeset[cnt];
+    else if (isdigit (codeset[cnt]))
+      *wp++ = codeset[cnt];
+  
+  *wp = '\0';
+
+  return retval;
+}
+
 void
 gtk_rc_init (void)
 {
-  gchar *locale_suffixes[3];
-  gint n_locale_suffixes = 0;
+  static gchar *locale_suffixes[8];
+  static gint n_locale_suffixes = 0;
+
   gint i, j;
-#ifdef HAVE_LC_MESSAGES
-  char *locale = setlocale (LC_MESSAGES, NULL);
-#else
-  char *locale = setlocale (LC_CTYPE, NULL);
-#endif
-  guint length;
-  char *p;
 
-  rc_style_ht = g_hash_table_new ((GHashFunc) gtk_rc_style_hash,
-				  (GCompareFunc) gtk_rc_style_compare);
-  pixmap_path[0] = NULL;
-  module_path[0] = NULL;
-  gtk_rc_append_default_pixmap_path();
-  gtk_rc_append_default_module_path();
+  static gboolean initted = FALSE;
 
-  gtk_rc_add_initial_default_files ();
-
-  if (strcmp (locale, "C") && strcmp (locale, "POSIX"))
+  if (!initted)
     {
-      /* Determine locale-specific suffixes for RC files
-       */
-      p = strchr (locale, '@');
-      length = p ? (p -locale) : strlen (locale);
+      char *locale = g_strdup (setlocale (LC_CTYPE, NULL));
+      char *p;
       
-      p = strchr (locale, '.');
-      if (p)
-	{
-	  locale_suffixes[n_locale_suffixes++] = g_strndup (locale, length);
-	  length = p - locale;
-	}
-      
-      p = strchr (locale, '_');
-      if (p)
-	{
-	  locale_suffixes[n_locale_suffixes++] = g_strndup (locale, length);
-	  length = p - locale;
-	}
+      initted = TRUE;
 
-      locale_suffixes[n_locale_suffixes++] = g_strndup (locale, length);
+      pixmap_path[0] = NULL;
+      module_path[0] = NULL;
+      gtk_rc_append_default_pixmap_path();
+      gtk_rc_append_default_module_path();
+      
+      gtk_rc_add_initial_default_files ();
+
+      if (strcmp (locale, "C") && strcmp (locale, "POSIX"))
+	{
+	  /* Determine locale-specific suffixes for RC files
+	   *
+	   * We normalize the charset into a standard form,
+	   * which has all '-' and '_' characters removed,
+	   * and is lowercase.
+	   *
+	   * the search is done in that order:
+	   * gtkrc.ll_cc.lowercasecodeset
+	   * gtkrc.ll_cc.normalizedcodeset
+	   * gtkrc.ll.lowercasecodeset
+	   * gtkrc.ll.normalizedcodeset
+	   * gtkrc.lowercasecodeset
+	   * gtkrc.normalizedcodeset
+	   * gtkrc.ll_cc
+	   * gtkrc.ll
+	   *
+	   */
+	  char *codeset = NULL;
+	  char *normalized_codeset = NULL;
+	  char *cc = NULL;
+	  char *ll;
+	  char *tmp = NULL;
+
+	  p = strchr (locale, '@');
+	  if (p)
+		  *p = '\0';
+
+#ifdef HAVE_CODESET
+	  codeset = nl_langinfo (CODESET);
+#endif
+	  p = strchr (locale, '.');
+	  if ((codeset==NULL) && (p))
+		  codeset = p+1;
+	  if (p)
+		  *p = '\0';
+	  
+	  if (codeset)
+	    {
+		/* if codeset is the result of nl_langinfo() it is
+		 * read-only so we need to copy it to a rw place */
+		tmp = g_strdup(codeset);
+		p = tmp;
+
+		while (*p)
+	          {
+		      /* tolower not used, because some locales are not
+		       * compatible with C locale in lowercasing ascii */
+		      if (*p >= 'A' && *p <= 'Z')
+			      *p = (*p)-'A'+'a';
+		      p++;
+		  }
+		codeset = tmp;
+		normalized_codeset =
+			_gtk_normalize_codeset(codeset, strlen(codeset));
+		if (strcmp(normalized_codeset,codeset) == 0)
+			normalized_codeset = NULL;
+	    }
+
+	  p = strchr (locale, '_');
+	  if (p) {
+		  cc = p+1;
+		  *p = '\0';
+	  }
+
+	  ll = locale;
+
+	  if (cc && codeset)
+	    locale_suffixes[n_locale_suffixes++] =
+	        g_strdup ( g_strconcat(ll,"_",cc,".",codeset,NULL));
+          if (cc && normalized_codeset)
+	    locale_suffixes[n_locale_suffixes++] =
+	        g_strdup ( g_strconcat(ll,"_",cc,".",normalized_codeset,NULL));
+	  if (codeset)
+            locale_suffixes[n_locale_suffixes++] =
+                g_strdup ( g_strconcat(ll,".",codeset,NULL));
+          if (normalized_codeset)
+	    locale_suffixes[n_locale_suffixes++] =
+	        g_strdup ( g_strconcat(ll,".",normalized_codeset,NULL));
+	  if (codeset)
+	    locale_suffixes[n_locale_suffixes++] =
+		g_strdup ( codeset );
+	  if (normalized_codeset)
+	    locale_suffixes[n_locale_suffixes++] =
+	        g_strdup ( normalized_codeset );
+	  if (cc)
+            locale_suffixes[n_locale_suffixes++] =
+	        g_strdup ( g_strconcat(ll,"_",cc,NULL));
+	  locale_suffixes[n_locale_suffixes++] = g_strdup ( ll );
+
+	  if (tmp) g_free (tmp);
+	}
+      g_free (locale);
     }
   
   i = 0;
@@ -514,6 +646,7 @@ gtk_rc_parse_file (const gchar *filename, gboolean reload)
   if (!lstat (rc_file->canonical_name, &statbuf))
     {
       gint fd;
+      GSList *tmp_list;
 
       rc_file->mtime = statbuf.st_mtime;
 
@@ -521,18 +654,19 @@ gtk_rc_parse_file (const gchar *filename, gboolean reload)
       if (fd < 0)
 	return;
 
-	{
-	  gint i;
-	  gchar *dir;
-	  
-	  dir = g_strdup(rc_file->canonical_name);
-	  for (i = strlen(dir) - 1; (i >= 0) && (dir[i] != '/'); i--)
-	    dir[i] = 0;
-	  gtk_rc_append_pixmap_path(dir);
-	  g_free(dir);
-	}
+      /* Temporarily push directory name for this file on
+       * a stack of directory names while parsing it
+       */
+      rc_dir_stack = g_slist_prepend (rc_dir_stack,
+				      g_dirname (rc_file->canonical_name));
       gtk_rc_parse_any (filename, fd, NULL);
 
+      tmp_list = rc_dir_stack;
+      rc_dir_stack = rc_dir_stack->next;
+
+      g_free (tmp_list->data);
+      g_slist_free_1 (tmp_list);
+
       close (fd);
     }
 }
@@ -550,12 +684,12 @@ gtk_rc_parse (const gchar *filename)
 GtkRcStyle *
 gtk_rc_style_new              (void)
 {
-  GtkRcStyle *new_style;
+  GtkRcStylePrivate *new_style;
 
-  new_style = g_new0 (GtkRcStyle, 1);
+  new_style = g_new0 (GtkRcStylePrivate, 1);
   new_style->ref_count = 1;
 
-  return new_style;
+  return (GtkRcStyle *)new_style;
 }
 
 void      
@@ -563,21 +697,62 @@ gtk_rc_style_ref (GtkRcStyle  *rc_style)
 {
   g_return_if_fail (rc_style != NULL);
 
-  rc_style->ref_count++;
+  ((GtkRcStylePrivate *)rc_style)->ref_count++;
+}
+
+/* Like g_slist_remove, but remove all copies of data */
+static GSList*
+gtk_rc_slist_remove_all (GSList   *list,
+			 gpointer  data)
+{
+  GSList *tmp;
+  GSList *prev;
+
+  prev = NULL;
+  tmp = list;
+
+  while (tmp)
+    {
+      if (tmp->data == data)
+	{
+	  if (list == tmp)
+	    list = list->next;
+
+	  if (prev) 
+	    prev->next = tmp->next;
+
+	  g_slist_free_1 (tmp);
+
+	  if (prev)
+	    tmp = prev->next;
+	  else
+	    tmp = list;
+	}
+      else
+	{
+	  prev = tmp;
+	  tmp = tmp->next;
+	}
+    }
+
+  return list;
 }
 
 void      
 gtk_rc_style_unref (GtkRcStyle  *rc_style)
 {
+  GtkRcStylePrivate *private = (GtkRcStylePrivate *)rc_style;
   gint i;
 
   g_return_if_fail (rc_style != NULL);
-  g_return_if_fail (rc_style->ref_count > 0);
+  g_return_if_fail (private->ref_count > 0);
 
-  rc_style->ref_count--;
+  private->ref_count--;
 
-  if (rc_style->ref_count == 0)
+  if (private->ref_count == 0)
     {
+      GSList *tmp_list1, *tmp_list2;
+	
       if (rc_style->engine)
 	{
 	  rc_style->engine->destroy_rc_style (rc_style);
@@ -595,16 +770,42 @@ gtk_rc_style_unref (GtkRcStyle  *rc_style)
 	if (rc_style->bg_pixmap_name[i])
 	  g_free (rc_style->bg_pixmap_name[i]);
       
-      g_free (rc_style);
-    }
-}
+      /* Now remove all references to this rc_style from
+       * realized_style_ht
+       */
+      tmp_list1 = private->rc_style_lists;
+      while (tmp_list1)
+	{
+	  GSList *rc_styles = tmp_list1->data;
+	  GtkStyle *style = g_hash_table_lookup (realized_style_ht, rc_styles);
+	  gtk_style_unref (style);
 
-static void
-gtk_rc_clear_realized_node (gpointer key,
-			    gpointer data,
-			    gpointer user_data)
-{
-  gtk_style_unref (data);
+	  /* Remove the list of styles from the other rc_styles
+	   * in the list
+	   */
+	  tmp_list2 = rc_styles;
+	  while (tmp_list2)
+	    {
+	      GtkRcStylePrivate *other_style = tmp_list2->data;
+
+	      if (other_style != private)
+		other_style->rc_style_lists =
+		  gtk_rc_slist_remove_all (other_style->rc_style_lists, rc_styles);
+		  
+	      tmp_list2 = tmp_list2->next;
+	    }
+
+	  /* And from the hash table itself
+	   */
+	  g_hash_table_remove (realized_style_ht, rc_styles);
+	  g_slist_free (rc_styles);
+
+	  tmp_list1 = tmp_list1->next;
+	}
+      g_slist_free (private->rc_style_lists);
+
+      g_free (private);
+    }
 }
 
 static void
@@ -642,13 +843,6 @@ gtk_rc_clear_styles (void)
       rc_style_ht = NULL;
     }
 
-  if (realized_style_ht)
-    {
-      g_hash_table_foreach (realized_style_ht, gtk_rc_clear_realized_node, NULL);
-      g_hash_table_destroy (realized_style_ht);
-      realized_style_ht = NULL;
-    }
-
   gtk_rc_free_rc_sets (gtk_rc_sets_widget);
   g_slist_free (gtk_rc_sets_widget);
   gtk_rc_sets_widget = NULL;
@@ -660,8 +854,6 @@ gtk_rc_clear_styles (void)
   gtk_rc_free_rc_sets (gtk_rc_sets_class);
   g_slist_free (gtk_rc_sets_class);
   gtk_rc_sets_class = NULL;
-
-  gtk_rc_init ();
 }
 
 gboolean
@@ -998,11 +1190,10 @@ gtk_rc_style_compare (const char *a,
 static GtkRcStyle*
 gtk_rc_style_find (const char *name)
 {
-  GtkRcStyle *rc_style;
-  
-  rc_style = g_hash_table_lookup (rc_style_ht, (gpointer) name);
-  
-  return rc_style;
+  if (rc_style_ht)
+    return g_hash_table_lookup (rc_style_ht, (gpointer) name);
+  else
+    return NULL;
 }
 
 /* Assumes ownership of rc_style */
@@ -1065,7 +1256,6 @@ gtk_rc_style_init (GSList *rc_styles)
   gint i;
 
   GtkStyle *style = NULL;
-  GtkRcStyle *proto_style;
 
   if (!realized_style_ht)
     realized_style_ht = g_hash_table_new ((GHashFunc)gtk_rc_styles_hash,
@@ -1075,13 +1265,16 @@ gtk_rc_style_init (GSList *rc_styles)
 
   if (!style)
     {
-      GSList *tmp_styles = rc_styles;
+      GtkRcStyle *proto_style;
+      GSList *tmp_styles;
       
       proto_style = gtk_rc_style_new ();
 
+      tmp_styles = rc_styles;
       while (tmp_styles)
 	{
 	  GtkRcStyle *rc_style = tmp_styles->data;
+	  GtkRcStylePrivate *rc_style_private;
 
 	  for (i=0; i<5; i++)
 	    {
@@ -1129,6 +1322,12 @@ gtk_rc_style_init (GSList *rc_styles)
 	      (proto_style->engine == rc_style->engine))
 	    proto_style->engine->merge_rc_style (proto_style, rc_style);
 
+	  /* Point from each rc_style to the list of styles */
+
+	  rc_style_private = (GtkRcStylePrivate *)rc_style;
+	  if (!g_slist_find (rc_style_private->rc_style_lists, rc_styles))
+	    rc_style_private->rc_style_lists = g_slist_prepend (rc_style_private->rc_style_lists, rc_styles);
+
 	  tmp_styles = tmp_styles->next;
 	}
 
@@ -1144,6 +1343,8 @@ gtk_rc_style_init (GSList *rc_styles)
 
       g_hash_table_insert (realized_style_ht, rc_styles, style);
     }
+  else
+    g_slist_free (rc_styles);
 
   return style;
 }
@@ -1218,7 +1419,7 @@ gtk_rc_parse_style (GScanner *scanner)
     return G_TOKEN_STRING;
   
   insert = FALSE;
-  rc_style = g_hash_table_lookup (rc_style_ht, scanner->value.v_string);
+  rc_style = gtk_rc_style_find (scanner->value.v_string);
   
   if (!rc_style)
     {
@@ -1250,7 +1451,7 @@ gtk_rc_parse_style (GScanner *scanner)
 	  return G_TOKEN_STRING;
 	}
       
-      parent_style = g_hash_table_lookup (rc_style_ht, scanner->value.v_string);
+      parent_style = gtk_rc_style_find (scanner->value.v_string);
       if (parent_style)
 	{
 	  for (i = 0; i < 5; i++)
@@ -1366,7 +1567,13 @@ gtk_rc_parse_style (GScanner *scanner)
     }
   
   if (insert)
-    g_hash_table_insert (rc_style_ht, rc_style->name, rc_style);
+    {
+      if (!rc_style_ht)
+	rc_style_ht = g_hash_table_new ((GHashFunc) gtk_rc_style_hash,
+					(GCompareFunc) gtk_rc_style_compare);
+      
+      g_hash_table_insert (rc_style_ht, rc_style->name, rc_style);
+    }
   
   return G_TOKEN_NONE;
 }
@@ -1503,26 +1710,49 @@ gtk_rc_parse_bg_pixmap (GScanner   *scanner,
   return G_TOKEN_NONE;
 }
 
+static gchar*
+gtk_rc_check_pixmap_dir (const gchar *dir, const gchar *pixmap_file)
+{
+  gchar *buf;
+  gint fd;
+
+  buf = g_strdup_printf ("%s%c%s", dir, '/', pixmap_file);
+  
+  fd = open (buf, O_RDONLY);
+  if (fd >= 0)
+    {
+      close (fd);
+      return buf;
+    }
+  
+  g_free (buf);
+
+  return NULL;
+}
+
 gchar*
 gtk_rc_find_pixmap_in_path (GScanner *scanner,
 			    const gchar *pixmap_file)
 {
   gint i;
-  gint fd;
-  gchar *buf;
+  gchar *filename;
+  GSList *tmp_list;
   
   for (i = 0; (i < GTK_RC_MAX_PIXMAP_PATHS) && (pixmap_path[i] != NULL); i++)
     {
-      buf = g_strdup_printf ("%s%c%s", pixmap_path[i], '/', pixmap_file);
+      filename = gtk_rc_check_pixmap_dir (pixmap_path[i], pixmap_file);
+      if (filename)
+	return filename;
+    }
+
+  tmp_list = rc_dir_stack;
+  while (tmp_list)
+    {
+      filename = gtk_rc_check_pixmap_dir (tmp_list->data, pixmap_file);
+      if (filename)
+	return filename;
       
-      fd = open (buf, O_RDONLY);
-      if (fd >= 0)
-	{
-	  close (fd);
-	  return buf;
-	}
-      
-      g_free (buf);
+      tmp_list = tmp_list->next;
     }
 
   if (scanner)
