@@ -54,7 +54,7 @@ struct _GtkThemeEnginePrivate {
 static GHashTable *engine_hash = NULL;
 
 GtkThemeEngine*
-gtk_theme_engine_get (gchar *name)
+gtk_theme_engine_get (const gchar *name)
 {
   GtkThemeEnginePrivate *result;
   
@@ -90,7 +90,7 @@ gtk_theme_engine_get (gchar *name)
        g_free(engine_path);
        if (!library)
 	 {
-	   g_warning (g_module_error());
+	   g_warning ("%s", g_module_error());
 	   return NULL;
 	 }
        else
@@ -107,7 +107,7 @@ gtk_theme_engine_get (gchar *name)
 		!g_module_symbol (library, "theme_exit", 
 				  (gpointer *)&result->exit))
 	      {
-		g_warning (g_module_error());
+		g_warning ("%s", g_module_error());
 		g_free (result);
 		return NULL;
 	      }
@@ -137,14 +137,17 @@ void
 gtk_theme_engine_unref (GtkThemeEngine *engine)
 {
   GtkThemeEnginePrivate *private;
+  private = (GtkThemeEnginePrivate *)engine;
 
   g_return_if_fail (engine != NULL);
+  g_return_if_fail (private->refcount > 0);
 
-  private = (GtkThemeEnginePrivate *)engine;
   private->refcount--;
 
   if (private->refcount == 0)
     {
+      private->exit();
+      
       g_hash_table_remove (engine_hash, private->name);
       
       g_module_close (private->library);
