@@ -77,6 +77,9 @@ typedef struct {
   GtkWindowLastGeometryInfo last;
 } GtkWindowGeometryInfo;
 
+#define GTK_WINDOW_HAS_FOCUS(window)                           \
+  ((window)->window_has_focus || (window)->window_has_pointer_focus)
+
 static void gtk_window_class_init         (GtkWindowClass    *klass);
 static void gtk_window_init               (GtkWindow         *window);
 static void gtk_window_set_arg            (GtkObject         *object,
@@ -455,7 +458,9 @@ gtk_window_set_default (GtkWindow *window,
     {
       if (window->default_widget)
 	{
-	  GTK_WIDGET_UNSET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
+	  if (window->focus_widget != window->default_widget ||
+	      !GTK_WIDGET_RECEIVES_DEFAULT (window->default_widget))
+	    GTK_WIDGET_UNSET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
 	  gtk_widget_draw_default (window->default_widget);
 	}
 
@@ -463,7 +468,9 @@ gtk_window_set_default (GtkWindow *window,
 
       if (window->default_widget)
 	{
-	  GTK_WIDGET_SET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
+	  if (window->focus_widget == NULL ||
+	      !GTK_WIDGET_RECEIVES_DEFAULT (window->focus_widget))
+	    GTK_WIDGET_SET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
 	  gtk_widget_draw_default (window->default_widget);
 	}
     }
@@ -852,12 +859,59 @@ gtk_window_finalize (GtkObject *object)
   GTK_OBJECT_CLASS(parent_class)->finalize (object);
 }
 
+
+static void
+reread_rc_files ()
+{
+  if (gtk_rc_reparse_all ())
+    {
+      /* If the above returned true, some of our RC files are out
+       * of date, so we need to reset all our widgets. Our other
+       * toplevel windows will also get the message, but by
+       * then, the RC file will up to date, so we have to tell
+       * them now.
+       */
+      GList *toplevels;
+
+      toplevels = gtk_container_get_toplevels();
+      while (toplevels)
+	{
+	  gtk_widget_reset_rc_styles (toplevels->data);
+	  toplevels = toplevels->next;
+	}
+    }
+}
+
 static void
 gtk_window_show (GtkWidget *widget)
 {
   GtkWindow *window = GTK_WINDOW (widget);
   GtkContainer *container = GTK_CONTAINER (window);
   gboolean need_resize;
+  GList *toplevels;
+  gboolean had_visible = FALSE;
+
+  /* If we have no windows shown at this point, then check for
+   * theme changes before showing the window. We really should
+   * be checking realized, not shown, but shown => realized,
+   * and checking in realize might cause reentrancy problems.
+   *
+   * Plus, this allows us to get the new size right before
+   * realizing.
+   */
+  toplevels = gtk_container_get_toplevels ();
+  while (toplevels)
+    {
+      if (GTK_WIDGET_VISIBLE (toplevels->data))
+	{
+	  had_visible = TRUE;
+	  break;
+	}
+      toplevels = toplevels->next;
+    }
+
+  if (!had_visible)
+    reread_rc_files ();
 
   GTK_WIDGET_SET_FLAGS (widget, GTK_VISIBLE);
 
@@ -922,6 +976,86 @@ gtk_window_hide (GtkWidget *widget)
     gtk_grab_remove (widget);
 }
 
+/* The algorithm used for window_has_pointer_focus is taken
+ * from the XTerm code.
+ */
+static GdkFilterReturn
+gtk_window_focus_filter (GdkXEvent *xevent,
+			 GdkEvent  *event,
+			 gpointer   data)
+{
+  GtkWindow *window = GTK_WINDOW (data);
+  XEvent *xev = (XEvent *)xevent;
+  GdkEvent extra_event;
+
+  switch (xev->xany.type)
+    {
+    case FocusIn:
+      {
+	switch (xev->xfocus.detail)
+	  {
+	  case NotifyAncestor:
+	  case NotifyNonlinear:
+	  case NotifyVirtual:
+	  case NotifyNonlinearVirtual:
+	    window->window_has_focus = TRUE;
+	    break;
+	  case NotifyPointer:
+	    window->window_has_pointer_focus = TRUE;
+	    break;
+	  case NotifyInferior:
+	  case NotifyPointerRoot:
+	  case NotifyDetailNone:
+	    break;
+	  }
+      }
+      break;
+    case FocusOut:
+      {
+	switch (xev->xfocus.detail)
+	  {
+	  case NotifyAncestor:
+	  case NotifyNonlinear:
+	  case NotifyVirtual:
+	  case NotifyNonlinearVirtual:
+	    window->window_has_focus = FALSE;
+	    break;
+	  case NotifyPointer:
+	    window->window_has_pointer_focus = FALSE;
+	    break;
+	  case NotifyInferior:
+	  case NotifyPointerRoot:
+	  case NotifyDetailNone:
+	    break;
+	  }
+      }
+      break;
+    case EnterNotify:
+    case LeaveNotify:
+      /* We only track the actual destination of keyboard events for real
+       * toplevels, not for embedded toplevels such as GtkPlug. The reason for
+       * this is that GtkPlug redirects events so the widget may effectively not
+       * have the focus even if it actually has the focus.
+       */
+      if (gdk_window_get_parent (GTK_WIDGET (window)->window) == GDK_ROOT_PARENT () &&
+	  xev->xcrossing.detail != NotifyInferior &&
+	  xev->xcrossing.focus && !window->window_has_focus)
+	{
+	  window->window_has_pointer_focus = (xev->xany.type == EnterNotify) ? TRUE : FALSE;
+
+	  extra_event.type = GDK_FOCUS_CHANGE;
+	  extra_event.focus_change.window = GTK_WIDGET (window)->window;
+	  extra_event.focus_change.send_event = FALSE;
+	  extra_event.focus_change.in = window->window_has_pointer_focus;
+
+	  gdk_event_put (&extra_event);
+	}
+      break;
+    }
+
+  return GDK_FILTER_CONTINUE;
+}
+
 static void
 gtk_window_map (GtkWidget *widget)
 {
@@ -939,6 +1073,8 @@ gtk_window_map (GtkWidget *widget)
       !GTK_WIDGET_MAPPED (window->bin.child))
     gtk_widget_map (window->bin.child);
 
+  gdk_window_add_filter (widget->window, gtk_window_focus_filter, widget);
+
   gdk_window_show (widget->window);
 }
 
@@ -951,7 +1087,7 @@ gtk_window_unmap (GtkWidget *widget)
   g_return_if_fail (GTK_IS_WINDOW (widget));
 
   GTK_WIDGET_UNSET_FLAGS (widget, GTK_MAPPED);
-  gdk_window_hide (widget->window);
+  gdk_window_withdraw (widget->window);
 
   window = GTK_WINDOW (widget);
   window->use_uposition = TRUE;
@@ -1349,6 +1485,7 @@ gtk_window_focus_out_event (GtkWidget     *widget,
   g_return_val_if_fail (event != NULL, FALSE);
 
   window = GTK_WINDOW (widget);
+
   if (window->focus_widget &&
       window->focus_widget != widget &&
       GTK_WIDGET_HAS_FOCUS (window->focus_widget))
@@ -1390,23 +1527,7 @@ gtk_window_read_rcfiles (GtkWidget *widget,
 	}
     }
 
-  if (gtk_rc_reparse_all ())
-    {
-      /* If the above returned true, some of our RC files are out
-       * of date, so we need to reset all our widgets. Our other
-       * toplevel windows will also get the message, but by
-       * then, the RC file will up to date, so we have to tell
-       * them now.
-       */
-      GList *toplevels;
-      
-      toplevels = gtk_container_get_toplevels();
-      while (toplevels)
-	{
-	  gtk_widget_reset_rc_styles (toplevels->data);
-	  toplevels = toplevels->next;
-	}
-    }
+  reread_rc_files ();
 }
 
 static gint
@@ -1455,52 +1576,47 @@ gtk_window_real_set_focus (GtkWindow *window,
   
   if (window->focus_widget)
     {
-      event.type = GDK_FOCUS_CHANGE;
-      event.window = window->focus_widget->window;
-      event.in = FALSE;
-      
       if (GTK_WIDGET_RECEIVES_DEFAULT (window->focus_widget) &&
 	  (window->focus_widget != window->default_widget))
         {
 	  GTK_WIDGET_UNSET_FLAGS (window->focus_widget, GTK_HAS_DEFAULT);
-	  /* if any widget had the default set there should be
-	     a default_widget, but might not so this is a sanity
-	     check */
+
 	  if (window->default_widget)
 	    GTK_WIDGET_SET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
         }
-	
-      gtk_widget_event (window->focus_widget, (GdkEvent*) &event);
+
+      if (GTK_WINDOW_HAS_FOCUS (window))
+	{
+	  event.type = GDK_FOCUS_CHANGE;
+	  event.window = window->focus_widget->window;
+	  event.in = FALSE;
+	  
+	  gtk_widget_event (window->focus_widget, (GdkEvent*) &event);
+	}
     }
   
   window->focus_widget = focus;
   
   if (window->focus_widget)
     {
-      event.type = GDK_FOCUS_CHANGE;
-      event.window = window->focus_widget->window;
-      event.in = TRUE;
+      if (GTK_WIDGET_RECEIVES_DEFAULT (window->focus_widget) &&
+	  (window->focus_widget != window->default_widget))
+	{
+	  if (GTK_WIDGET_CAN_DEFAULT (window->focus_widget))
+	    GTK_WIDGET_SET_FLAGS (window->focus_widget, GTK_HAS_DEFAULT);
 
-      if (window->default_widget)
-        {
-          if (GTK_WIDGET_RECEIVES_DEFAULT (window->focus_widget) &&
-	      (window->focus_widget != window->default_widget))
-            {
-	      if (GTK_WIDGET_CAN_DEFAULT (window->focus_widget))
-	        GTK_WIDGET_SET_FLAGS (window->focus_widget, GTK_HAS_DEFAULT);
-	      GTK_WIDGET_UNSET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
-            }
-	  else
-	    {
-	      GTK_WIDGET_SET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
-	    }
+	  if (window->default_widget)
+	    GTK_WIDGET_UNSET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
 	}
       
-      gtk_widget_event (window->focus_widget, (GdkEvent*) &event);
-    }
-  else if (window->default_widget)
-    {
-      GTK_WIDGET_SET_FLAGS (window->default_widget, GTK_HAS_DEFAULT);
+      if (GTK_WINDOW_HAS_FOCUS (window))
+	{
+	  event.type = GDK_FOCUS_CHANGE;
+	  event.window = window->focus_widget->window;
+	  event.in = TRUE;
+
+	  gtk_widget_event (window->focus_widget, (GdkEvent*) &event);
+	}
     }
   
   if (window->default_widget &&
